import typing

from numpy.typing import ArrayLike
import pandas as pd

from ..utils import models, constants

# from ..utils import plot


class AggregatedData:
    """A class to represent aggregated data of a traffic measurement station.

    Attributes:
    -----------
    data : typing.Optional[pd.DataFrame]
        Aggregated data of the traffic measurement station.
    aggregation_time_period : int
        Aggregation time period value, by default DEF_AGG_TIME_PER
    aggregation_time_period_unit : str
        Aggregation time period unit, either "min" or "sec", by default "min"
    models : models.ModelResults
        Estimated models for the aggregated data.
    quantiles : list[float]
        List of quantiles for which the models were estimated.

    Methods:
    --------
    estimate_model(model_type: str, quantiles: list = None, penalty: str = None, eta: float = None, email: str = None) -> None
        Estimates a convex nonparametric least squares (CNLS) or convex quantile regression (CQR) model for the\\
        aggregated data.
    plot(quantiles=None, figure_name=None) -> None
        Plot estimated quantiles of the fundamental diagram.
    """  # noqa E501

    def __init__(self):
        self.data: typing.Optional[pd.DataFrame] = None
        self._flow: typing.Optional[ArrayLike] = None
        self._speed: typing.Optional[ArrayLike] = None
        self._density: typing.Optional[ArrayLike] = None
        self.aggregation_time_period: int = constants.DEF_AGG_TIME_PER
        self.aggregation_time_period: str = "min"
        self.models: models.ModelResults = models.ModelResults()
        self.quantiles: list[float] = []
        pass

    # def load_csv(
    #    self,
    #    path: str,
    #    separator: str = ",",
    #    column_names: list[str] = None,
    #    aggregation_time: int = 5,
    # ) -> None:
    #    """Load aggregated data from a csv file.

    #    Parameters
    #    ----------
    #    path : str
    #        Path to the file to be loaded.
    #    separator : str, optional
    #        Separator used in the file, by default ","
    #    column_names : list[str], optional
    #        Names of columns in the file. Make sure to include "density" and "flow" columns,
    #        by default ["density", "flow"]
    #    aggregation_time : int, optional
    #        Aggregation time period used in the file, by default 5

    #    Raises
    #    ------
    #    FileNotFoundError
    #        File was not found on the given `path`.
    #    """
    #    if column_names is None:
    #        column_names = ["density", "flow"]

    #    path = pathlib.Path(path)

    #    if not path.exists():
    #        raise FileNotFoundError(f"[LOG] File {path} does not exist.")
    #    else:
    #        assert "density" in column_names, "Column names must contain 'density'."
    #        assert "flow" in column_names, "Column names must contain 'flow'."

    #        self.data = pd.read_csv(path, header=0, sep=separator, names=column_names)
    #        self._flow = self.data["flow"].to_numpy()
    #        self._density = self.data["density"].to_numpy()
    #        self.aggregation_time = aggregation_time
    #    pass

    def estimate_model(
        self,
        model_type: str,
        quantiles: typing.Optional[list[float]] = None,
        penalty: typing.Optional[str] = None,
        eta: typing.Optional[float] = None,
        email: str = None,
    ) -> None:
        """ Estimates a convex nonparametric least squares (CNLS) or convex quantile regression (CQR) model for the\\
            aggregated data.

        Parameters
        ----------
        model_type : str
            Type of the model to be estimated. Takes value `'quantile'` for the convex quantile regression (CQR)
            model or `'mean'` for the convex nonparametric least squares (CNLS) model
        quantiles : list[float]
            Quantiles to be estimated for the `model_type=quantile`
        penalty : str, optional
            Type of penalty to be used in the model. Takes values `penalty=None` (no penalty), `penalty='l1'` (L1 norm),
            `penalty='l2'` (L2 norm), or `penalty='l3'` (Lipschitz norm), by default None
        eta : float, optional
            Value of the tuning parameter if `penalty` is in `['l1', 'l2', 'l3']`, by default None
        email : str, optional
            For external optimization on the NEOS server, by default None

        Raises
        ------
        AssertionError
            Model type must be "quantile" or "mean".
            Quantiles must be a non-empty list.
            Quantiles must be between 0 and 1.
            Data must be loaded first.
            Data contains more than 3000 observations. Apply bagging first.
        """
        # if quantiles is None:
        #    quantiles = [0.5]

        assert model_type in [
            "quantile",
            "mean",
        ], "[LOG] AssertionError:Model type must be 'quantile' or 'mean'."
        assert self.data is not None, "[LOG] AssertionError: No data is available"
        assert (
            len(self._flow) <= 3000
        ), "[LOG] AssertionError:Data contains more than 3000 observations. Apply bagging first."
        if penalty is not None:
            assert penalty in [
                "l1",
                "l2",
                "l3",
            ], "[LOG] AssertionError:Penalty must be None, 'l1', 'l2' or 'l3'."
            assert (
                eta is not None
            ), "[LOG] AssertionError: eta must be specified if penalty is selected."

        if model_type == "quantile":
            assert (
                len(quantiles) > 0
            ), "[LOG] AssertionError: Quantiles must be a non-empty list."
            assert all(
                [0 < q < 1 for q in quantiles]
            ), "[LOG] AssertionError: Quantiles must be between 0 and 1."

            [self.quantiles.append(x) for x in quantiles if x not in self.quantiles]
            self.quantiles.sort()

            for q in quantiles:
                if penalty is not None:
                    model = models._pcqr(
                        x=self._density,
                        y=self._flow,
                        quantile=q,
                        penalty=penalty,
                        eta=eta,
                        email=email,
                    )
                else:
                    model = models._cqr(
                        x=self._density, y=self._flow, quantile=q, email=email
                    )
                self.models._add_model(model, q, penalty, eta)

        elif model_type == "mean":
            if penalty is not None:
                model = models._pcnls(
                    x=self._density,
                    y=self._flow,
                    penalty=penalty,
                    eta=eta,
                    email=email,
                )
            else:
                model = models._cnls(x=self._density, y=self._flow, email=email)
            self.models._add_model(model, "mean", penalty, eta)
        pass


#    def plot(self, quantiles=None, figure_name=None) -> None:
#        """
#        Plot estimated quantiles of the fundamental diagram.
#        Parameters
#        ----------
#        quantiles : list[float] or None, optional
#            List of quantiles to be plotted. If `None`, all estimated quantiles will be plotted, by default None
#        figure_name : str or None, optional
#            Name of the figure is specified when the figure is intended to be saved, by default None
#        """
#        plot.plot(
#            self, representation="agg", quantiles=quantiles, figure_name=figure_name
#        )
#        pass



import typing
import pandas as pd
import numpy as np

from .traffic_data_classes.bagged_data import BaggedData

from .traffic_data_classes.raw_data import RawData
from .traffic_data_classes.aggregated_data import AggregatedData
from .utils import load, process
from .utils.constants import DEF_AGG_TIME_PER, DEF_NUM_BAGS_DENSITY, DEF_NUM_BAGS_FLOW


class TrafficMeasurementStation:
    class Raw(RawData):
        """A class to represent raw data of a traffic measurement station.

        Attributes
        ----------
        tms_id : int
            Traffic measurement station ID, learn more at \
            `Digitraffic <https://www.digitraffic.fi/en/road-traffic/lam/>`_

        days_list : list[tuple[int, int]]
            List of tuples containing year and ordinal day of the year

        data_raw : typing.Optional[pd.DataFrame]
            Raw data of the traffic measurement station

        data_processed : typing.Optional[pd.DataFrame]
            Processed data of the traffic measurement station

        cleaning_params : typing.Optional[dict]
            Parameters used for cleaning the data
        """

        def __init__(self, tms_id: int, days_list: list[tuple[int, int]]):
            super().__init__()
            self.tms_id: int = tms_id
            self.days_list: list[tuple[int, int]] = days_list

            self.data_raw: typing.Optional[pd.DataFrame] = None
            self.data_processed: typing.Optional[pd.DataFrame] = None
            self.cleaning_params: typing.Optional[dict] = None
            pass

        def load(
            self,
            load_path: typing.Optional[str] = None,
            save: bool = False,
            save_path: typing.Optional[str] = None,
            sort: bool = True,
        ) -> None:
            """Load data from local storage or from the Internet.

            Parameters
            ----------
            load_path : typing.Optional[str], optional
                Path to load data from local storage, by default None
            save : bool, optional
                Whether to save the loaded data to local storage, by default False
            save_path : typing.Optional[str], optional
                Path to save the loaded data to local storage, by default None
            sort : bool, optional
                Whether to sort the loaded data by `total_time`, by default True
            """
            self.data_raw = load.read_many_reports(
                self.tms_id,
                self.days_list,
                save=save,
                save_path=save_path,
                sort_total_time=sort,
                load_path=load_path,
            )
            pass

        def clean(self, direction: int, **cleaning_params) -> None:
            """Clean the loaded data.

            Parameters
            ----------
            direction : int
                Direction of the traffic for a given traffic measurement station.
                It is either 1 or 2. Check the direction of a traffic measurement station at
                `Finntraffic <https://liikennetilanne.fintraffic.fi/kartta/>`_
            cleaning_params : dict
                Parameters for cleaning the data, check :func:`~roadtraffic.utils.load.process_data`
            """
            assert (
                self.data_raw is not None
            ), "[LOG] AssertionError: load the data first."
            assert direction in [
                1,
                2,
            ], "[LOG] AssertionError: direction must be either 1 or 2."
            self.cleaning_params = cleaning_params

            try:
                lanes = cleaning_params["lanes"]
            except KeyError:
                lanes = None

            try:
                delete_if_faulty = cleaning_params["delete_if_faulty"]
            except KeyError:
                delete_if_faulty = True

            try:
                hour_from = cleaning_params["hour_from"]
            except KeyError:
                hour_from = 0

            try:
                hour_to = cleaning_params["hour_to"]
            except KeyError:
                hour_to = 23

            try:
                assign_date = cleaning_params["assign_date"]
            except KeyError:
                assign_date = True

            try:
                calculate_vehicles = cleaning_params["calculate_vehicles"]
            except KeyError:
                calculate_vehicles = True

            self.data_processed = load.process_data(
                self.data_raw,
                direction=direction,
                lanes=lanes,
                delete_if_faulty=delete_if_faulty,
                hour_from=hour_from,
                hour_to=hour_to,
                assign_date=assign_date,
                calculate_vehicles=calculate_vehicles,
            )
        pass

    class Aggregated(AggregatedData):
        """A class to represent aggregated data of a traffic measurement station."""

        def __init__(self):
            super().__init__()
            pass

    class Bagged(BaggedData):
        """A class to represent bagged data of a traffic measurement station."""

        def __init__(self):
            super().__init__()
            pass

    def __init__(self, tms_id: int, days_list: list[tuple[int, int]]):
        self.tms_id: int = tms_id
        self.days_list: list[tuple[int, int]] = days_list

        self.raw = TrafficMeasurementStation.Raw(self.tms_id, self.days_list)
        self.agg = TrafficMeasurementStation.Aggregated()
        self.bag = TrafficMeasurementStation.Bagged()
        pass

    def raw_to_agg(
        self,
        aggregation_time_period: int = DEF_AGG_TIME_PER,
        aggregation_time_period_unit: str = "min",
    ) -> None:
        """Convert raw data to aggregated data.

        Parameters
        ----------
        aggregation_time_period : int
            Aggregation time period value, by default DEF_AGG_TIME_PER
        aggregation_time_period_unit : str
            Aggregation time period unit, either "min" or "sec", by default "min"

        Raises
        ------
        AssertionError
            If raw data is not cleaned.
        """
        assert (
            self.raw.data_processed is not None
        ), "[LOG] AssertionError: clean raw data first."
        self.agg.data = process.aggregate(
            self.raw.data_processed,
            aggregation_time_period=aggregation_time_period,
            aggregation_time_period_unit=aggregation_time_period_unit,
        )
        self.agg.aggregation_time_period = aggregation_time_period
        self.agg.aggregation_time_period_unit = aggregation_time_period_unit

        _density = self.agg.data["density"].to_numpy().T
        _flow = self.agg.data["flow"].to_numpy().T
        _speed = self.agg.data["period_speed"].to_numpy().T
        _data_array = (np.stack([_density, _flow, _speed], axis=0)).T
        _data_array = _data_array[np.argsort(_data_array[:, 0])].T
        self.agg._density = _data_array[0]
        self.agg._flow = _data_array[1]
        self.agg._speed = _data_array[2]

        pass

    def agg_to_bag(
        self,
        num_bags_density: int = DEF_NUM_BAGS_DENSITY,
        num_bags_flow: int = DEF_NUM_BAGS_FLOW,
    ) -> None:
        """Convert aggregated data to bagged data.

        Parameters
        ----------
        num_bags_density : int
            Number of bags for density to be used in the bagging procedure, by default DEF_NUM_BAGS_DENSITY
        num_bags_flow : int
            Number of bags for flow to be used in the bagging procedure, by default DEF_NUM_BAGS_FLOW

        Raises
        ------
        AssertionError
            If aggregated data is not aggregated.
        """
        assert (
            self.agg.data is not None
        ), "[LOG] AssertionError: aggregate data first."

        self.bag.data = process.bagging(
            self.agg.data, grid_size_x=num_bags_density, grid_size_y=num_bags_flow
        )
        self.bag.bag_size_flow = num_bags_flow
        self.bag.bag_size_density = num_bags_density

        _density = self.bag.data["centroid_density"].to_numpy().T
        _flow = self.bag.data["centroid_flow"].to_numpy().T
        _weight = self.bag.data["weight"].to_numpy().T
        _data_array = (np.stack([_density, _flow, _weight], axis=0)).T
        _data_array = _data_array[np.argsort(_data_array[:, 0])].T
        self.bag._density = _data_array[0]
        self.bag._flow = _data_array[1]
        self.bag._weight = _data_array[2]

        pass



from roadtraffic.fintraffic import TrafficMeasurementStation
import roadtraffic.utils.funcs as funcs


tms = TrafficMeasurementStation(146, [(2019, 270), (2019, 271)])
tms.raw.load()
tms.raw.clean(direction=1, lanes=[1, 2, 3])
tms.raw_to_agg()
tms.agg_to_bag()


def test_cnls():
    tms.agg.estimate_model("mean", email="test1@test.test")
    assert tms.agg.models.get_problem_status("mean") == 1
    assert len(tms.agg.models.get_estimate("mean")) == 576
    assert len(tms.agg.models.get_frontier("mean")) == 576
    assert tms.agg.models.get_coefficients("mean").shape == (576, 2)
    assert len(tms.agg.models.get_models()) == 1


def test_pcnls():
    tms.agg.estimate_model("mean", penalty="l1", eta=10000, email="test2@test.test")
    assert tms.agg.models.get_problem_status("mean", penalty="l1", eta=10000) == 1


def test_cqr():
    tms.agg.estimate_model("quantile", quantiles=[0.5], email="test3@test.test")
    assert tms.agg.models.get_problem_status(quantile=0.5) == 1


def test_pcqr():
    tms.agg.estimate_model(
        "quantile", quantiles=[0.5], penalty="l1", eta=10000, email="test4@test.test"
    )
    assert (
        tms.agg.models.get_problem_status(
            quantile=0.5, penalty="l1", eta=10000
        )
        == 1
    )


def test_wcnls():
    tms.bag.estimate_model("mean", email="test5@test.test")
    assert tms.bag.models.get_problem_status("mean") == 1


def test_pwcnls():
    tms.bag.estimate_model("mean", penalty="l1", eta=10000, email="test6@test.test")
    assert tms.bag.models.get_problem_status("mean", penalty="l1", eta=10000) == 1


def test_wcqr():
    tms.bag.estimate_model("quantile", quantiles=[0.5], email="test7@test.test")
    assert tms.bag.models.get_problem_status(quantile=0.5) == 1


def test_pwcqr():
    tms.bag.estimate_model(
        "quantile", quantiles=[0.5], penalty="l1", eta=10000, email="test8@test.test"
    )
    assert (
        tms.bag.models.get_problem_status(
            quantile=0.5, penalty="l1", eta=10000
        )
        == 1
    )

def test_date_to_day():
    assert funcs.date_to_day(funcs.day_to_date(2019,270)) == (2019,270)



# import dependencies
import datetime
import os
import pathlib
import time
import typing
import warnings

import pandas as pd
import requests
from tqdm import tqdm

from .constants import (
    DEF_COL_NAMES_FINTRAFFIC,
    URL_FINTRAFFIC,
    DEF_FILENAME_U,
)


# Download the .csv report of `tms_id` station for the day `day_of_year`
def read_raw_report(
    tms_id: int,
    year: int,
    day: int,
    load_path: typing.Optional[str] = None,
    save: bool = False,
    save_path: typing.Optional[str] = None,
    sort_total_time: bool = True,
) -> pd.DataFrame:
    """
    Download the raw data from Fintraffic for the `tms_id` station for the `day` of the `year`. \
    By default the data is cleaned - faulty observations are deleted. \
    The type of vehicle is determined based on the classification from Fintraffic.

    | It is possible to save a `.pkl` file, setting `save` to True. The default filename format "lamraw_TMS_YY_DD_bz2.pkl" is used. 
    | It is possible to locally read a `.pkl` file locally, specifying a `path_name`. \
        At first, the search for the processed files is done. \
        If it is not found, then the unprocessed file is looked up. \
        Processed file is the one load with the use of this function. \
        Unprocessed file is the one loaded using some other way. 

    Parameters
    ----------
    tms_id : int
        The identity number of a traffic measurement station (TMS). \
        Meta-data about TMS is available `here <https://www.digitraffic.fi/en/road-traffic/#current-data-from-tms-stations>`_.
    year : int
        The year data was collected in the 4-digit format. Data is available starting from 1995 only. 
    day : int
        The day of the year the data was collected. \
        The day is provide as an integer in range(1, 366), with 1 - January 1st, \
        365 (366 in leap year) - December 31st. \
        To caclulate the day of the year from a date you can use `roadtraffic.utils.funcs.date_to_day()` function.
    save : bool, optional
        If specified, saves loaded data in the .pkl file format with bz2-compression with the default filename format "lamraw_TMS_YY_DD_bz2_u.pkl", by default False
    load_path : str, optional
        If specified, reads data from the `path_name` folder from a locally saved bz2-compressed .pkl file. \
        Unprocessed "lamraw_TMS_YY_DD_bz2_u.pkl" files can be read, by default None
    save_path : str, optional
        If specified, saves loaded data in the `save_path` folder in the .pkl file format with bz2-compression with \
        the default filename format "lamraw_TMS_YY_DD_bz2_u.pkl", by default None
    sort_total_time : bool, optional
        If specified, sorts the data by the `total_time` column, by default True

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the raw data.
        
    Raises
    ------
    AssertionError
        The day is incorrect, it should be 1 < day < 366.
    AssertionError
        The data is available starting from year 1995 only.

    See Also
    --------
    :func:`roadtraffic.utils.funcs.date_to_day`
        A function to convert a date into day of the year. 
    """  # noqa E501
    # Initiate timer
    # start_time = time.perf_counter()

    # Check assert errors
    # fmt: off
    assert 1 <= day <= 366, "[LOG] AssertionError: the day is incorrect, it should be 1 < day < 366"  # noqa
    assert 1995 <= year, "[LOG] AssertionError: the data is available starting from year 1995 only "  # noqa
    # fmt: on

    # Create column names for the pd.DataFrame
    column_names = DEF_COL_NAMES_FINTRAFFIC

    # Initiate an empty pd.DataFrame
    df = pd.DataFrame()

    # Create the actual filename for processed data
    # file_name = DEF_FILENAME

    # file_name = (
    #    file_name.replace("TMS", str(tms_id))
    #    .replace("YY", str(year)[2:4])
    #    .replace("DD", str(day))
    # )

    # Create the actual filename for unprocessed data
    file_name_u = DEF_FILENAME_U

    file_name_u = (
        file_name_u.replace("TMS", str(tms_id))
        .replace("YY", str(year)[2:4])
        .replace("DD", str(day))
    )

    if load_path is not None:

        file_path_u = load_path + "/" + file_name_u
        file_path_u = pathlib.Path(file_path_u)

        if os.path.isfile(file_path_u):
            df = pd.read_pickle(file_path_u, compression="bz2")

        if df.empty:
            # Stop timer
            # end_time = time.perf_counter()
            message = f"[LOG] Warning: {file_name_u} is not found at {load_path}. An empty pd.DataFrame will be returned." # noqa E501
            warnings.warn(message=message)
            if save is True:
                warnings.warn(
                    "[LOG] Warning: It is impossible to save a file, as data was not loaded. Check warning above."
                    # noqa E501
                )
        else:

            # Stop timer
            # end_time = time.perf_counter()

            # print(
            #    f"File {file_name_u} is loaded. "
            #    + f"Time spent: {end_time - start_time:.4f} sec."  # noqa E501
            # )

            # Save to .pkl if necessary
            if save is True:
                if save_path is None:
                    s_path = file_name_u
                else:
                    s_path = os.path.join(save_path, file_name_u)
                df.to_pickle(s_path, compression="bz2")
                print(f"[LOG] Data is successfully saved to {s_path}")

    else:
        # Assign URL path for data loading
        url = URL_FINTRAFFIC

        # Create the actual url
        url = (
            url.replace("TMS", str(tms_id))
            .replace("YY", str(year)[2:4])
            .replace("DD", str(day))
        )

        # Call the link to receive the response
        r = requests.get(url)

        # Try to download the file
        if r.status_code != 404:
            # Check that there is no connection limit
            if r.status_code == 429:
                for i in range(59, 0, -1):
                    print(
                        f"[LOG] Too many requests, waiting {i} seconds to connect again",
                        end="\x1b[1K\r",
                        flush=True,
                    )
                    time.sleep(1)
                r = requests.get(url)
            if r.status_code == 429:
                message = "[LOG] The server is overloaded. Please, try to load the data later. An empty pd.DataFrame will be returned"  # noqa E501
                warnings.warn(message=message)
                return pd.DataFrame()
            elif r.status_code == 404:
                # Stop timer
                # end_time = time.perf_counter()
                # print(f"Time spent: {end_time - start_time:.4f} seconds")
                message = f"[LOG] Warning: file at {url} does not exist. Try to select another day. An empty pd.DataFrame will be returned."  # noqa E501
                warnings.warn(message=message)
                if save is True:
                    warnings.warn(
                        "[LOG] Warning: It is impossible to save a file, as data was not loaded. Check warning above."
                        # noqa E501
                    )
            else:
                with open(file_name_u[:-4] + ".csv", "wb") as f:
                    f.write(r.content)
                df = pd.read_csv(
                    file_name_u[:-4] + ".csv",
                    delimiter=";",
                    names=column_names,
                )
                os.remove(file_name_u[:-4] + ".csv")

                # Stop timer
                # end_time = time.perf_counter()
                # print(
                #    f"Download successful - file for the sensor {tms_id} for the day {day} in year {year} was loaded in {end_time - start_time:.4f} seconds"  # noqa E501
                #)  # noqa E501

                # Save to .pkl if necessary
                if save is True:
                    if save_path is None:
                        s_path = file_name_u
                    else:
                        s_path = os.path.join(save_path, file_name_u)
                    df.to_pickle(s_path, compression="bz2")
                    print(f"[LOG] Data is successfully saved to {s_path}")
        else:
            # Stop timer
            # end_time = time.perf_counter()
            # print(f"Time spent: {end_time - start_time:.4f} seconds")
            message = f"[LOG] Warning: file at {url} does not exist. Try to select another day. An empty pd.DataFrame will be returned."  # noqa E501
            warnings.warn(message=message)
            if save is True:
                warnings.warn(
                    "[LOG] Warning: It is impossible to save a file, as data was not loaded. Check warning above."
                    # noqa E501
                )

    if not df.empty and sort_total_time:
        df.sort_values(by=["total_time"], inplace=True, ignore_index=True)
    return df


def read_many_reports(
        tms_id: int,
        days_list: list[tuple[int, int]],
        save: bool = False,
        load_path: typing.Optional[str] = None,
        save_path: typing.Optional[str] = None,
        sort_total_time: bool = True
) -> pd.DataFrame:
    """
    Download the raw data from Fintraffic for the specified station for a list of days.

    Parameters
    ----------
    tms_id: int
        The identity number of a traffic measurement station (TMS).
    days_list: list
        A list of tuples to download the data for. A format for each day is (year, day of year).
    load_path:
        If specified, reads data from the `path_name` folder from a locally saved bz2-compressed .pkl file. \
        Unprocessed "lamraw_TMS_YY_DD_bz2_u.pkl" files can be read, by default None
    save: bool, optional
        If specified, saves loaded data in the .pkl file format with bz2-compression with filename defined by
        `save_filename`, by default False
    save_path:
        If specified, saves data for each loaded day in the `save_path` folder in the .pkl file format with
        bz2-compression with a standard filename "lamraw_TMS_YY_DD_bz2_u.pkl", by default None
    sort_total_time : bool, optional
        If specified, sorts the data for each day by the `total_time` column and resets index , by default True

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the raw data.
    """
    assert isinstance(days_list, list), "days_list must be a list."
    assert all([isinstance(day, tuple) for day in days_list]), "[LOG] AssertionError: days_list must contain only tuples." # noqa E501
    assert all([isinstance(day[0], int) for day in days_list]), "[LOG] AssertionError: days_list must contain only tuples of integers." # noqa E501
    assert all([isinstance(day[1], int) for day in days_list]), "[LOG] AssertionError: days_list must contain only tuples of integers." # noqa E501

    # Initiate timer
    start_time = time.perf_counter()
    total_files = len(days_list)
    loaded_files = 0

    # Initiate an empty pd.DataFrame
    df_all = pd.DataFrame()

    # Iterate over the list of days
    for year, day in tqdm(days_list, desc="Loading files", unit="files"):
        df = read_raw_report(
            tms_id=tms_id,
            year=year,
            day=day,
            load_path=load_path,
            save=save,
            save_path=save_path,
            sort_total_time=sort_total_time,
        )
        if not df.empty:
            if df_all.empty:
                df_all = df
            else:
                df_all = pd.concat([df_all, df], ignore_index=True)
            loaded_files += 1

    # Stop timer
    end_time = time.perf_counter()
    print(
        f"[LOG] Download finished - {loaded_files}/{total_files} files were loaded in {end_time - start_time:.4f} seconds" # noqa E501
    )

    return df_all


def process_data(
    df: pd.DataFrame,
    direction: typing.Optional[int] = None,
    lanes: typing.Optional[list] = None,
    delete_if_faulty: bool = True,
    hour_from: int = 0,
    hour_to: int = 23,
    assign_date: bool = True,
    calculate_vehicles: bool = True,
) -> pd.DataFrame:
    """Process raw data.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be cleaned.
    direction : int, optional
        Direction of the traffic movement, can take values 1 or 2 only. Check your TMS for the direction,
        by default None
    lanes : list, optional
        Lanes to be included in the data, by default None
    delete_if_faulty : bool, optional
        If faulty data should be excluded, by default True
    hour_from : int, optional
        Hour from which beginning the data should be included, by default 0
    hour_to : int, optional
        Hour to which end the data should be included, by default 23
    assign_date : bool, optional
        If `True`, assigns the date to the data, by default True
    calculate_vehicles : bool, optional
        If `True`, calculates the number of different vehicles, by default True

    Raises
    ------
    AssertionError
        direction must be 1 or 2.
    AssertionError
        lanes must be a list.
    AssertionError
        lanes must contain only integers.
    AssertionError
        hour_from must be greater or equal to 0 but less than 23.
    AssertionError
        hour_to must be greater than hour_from and less than 24.

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the processed data. Index is reset.
    """
    # DIRECTION
    if direction is not None:
        assert direction in [1, 2], "[LOG] AssertionError: direction must be 1 or 2."
        df = df[df["direction"] == direction]

    # LANES
    if lanes is not None:
        assert lanes is None or isinstance(lanes, list), "[LOG] AssertionError: lanes must be a list."
        assert lanes is None or all(
            [isinstance(lane, int) for lane in lanes]
        ), "[LOG] AssertionError: lanes must contain only integers."

        df = df[df["lane"].isin(lanes)]

    # HOUR_TO and HOUR_FROM
    assert (
        0 <= hour_from <= 23
    ), "[LOG] AssertionError: hour_from must be greater or equal to 0 but less or equal to 23."
    assert (
        hour_from <= hour_to < 24
    ), "[LOG] AssertionError: hour_to must be greater or equal to hour_from and less than 24."
    df = df[(df["hour"] >= hour_from) & (df["hour"] <= hour_to)]

    # IF_FAULTY
    if delete_if_faulty:
        df = df[df["faulty"] == 0]

    # Assign dates
    if assign_date:
        df["temp"] = df.year.astype("str") + "-" + df.day.astype("str")
        df["date"] = df.temp.apply(
            lambda x: datetime.date(int(x.split("-")[0]) + 2000, 1, 1)
            + datetime.timedelta(int(x.split("-")[1]) - 1)
        )
        df.drop(columns=['temp'], inplace=True)

    # Calculate the number of different vehicles
    if calculate_vehicles:
        df["cars"] = df["vehicle"].apply(lambda x: 1 if x == 1 else 0)
        df["buses"] = df["vehicle"].apply(lambda x: 1 if x == 3 else 0)
        df["trucks"] = df["vehicle"].apply(
            lambda x: 1 if x == 2 or x == 4 or x == 5 or x == 6 or x == 7 else 0
        )

    df.reset_index(drop=True, inplace=True)

    return df



from typing import Union, Any

URL_FINTRAFFIC: str = (
    "https://tie.digitraffic.fi/api/tms/v1/history/raw/lamraw_TMS_YY_DD.csv"
)
"""
URL to access raw data from traffic measurement stations, \
installed on the roads in Finland.

Source: `Finntraffic <https://www.fintraffic.fi/en>`_ \
/ `Digitraffic <https://www.digitraffic.fi/en/>`_ under \
`CC 4.0 BY <https://creativecommons.org/licenses/by/4.0/>`_ license.
"""

DEF_COL_NAMES_FINTRAFFIC: list[Union[str, Any]] = [
    "id",
    "year",
    "day",
    "hour",
    "minute",
    "second",
    "hund_second",
    "length",
    "lane",
    "direction",
    "vehicle",
    "speed",
    "faulty",
    "total_time",
    "time_interval",
    "queue_start",
]
"""
Standardized column names for files containing traffic data from Finland.
"""

DEF_AGG_TIME_PER: int = 5
"""
Default aggregation time period.
"""

DEF_TAU_LIST: list[float] = [0.5]
"""
Default list of quantiles to be estimated.
"""

DEF_FILENAME: str = "lamraw_TMS_YY_DD_bz2.pkl"
"""
Default filename to load processed traffic data from local file. \
Processing is done with :func:`~roadtraffic.utils.load.read_report`
"""

DEF_FILENAME_U: str = "lamraw_TMS_YY_DD_bz2_u.pkl"
"""
Default filename to load raw data from local file.
"""

DEF_NUM_BAGS_DENSITY: int = 70
"""
Default number of bags for density to be used in the bagging procedure.
"""

DEF_NUM_BAGS_FLOW: int = 400
"""
Default number of bags for flow to be used in the bagging procedure.
"""



import typing

from numpy.typing import ArrayLike
import pandas as pd

from ..utils import models


class BaggedData:
    """A class to represent bagged data of a traffic measurement station."""

    def __init__(self):
        self.data: typing.Optional[pd.DataFrame] = None
        self._flow: typing.Optional[ArrayLike] = None
        self._density: typing.Optional[ArrayLike] = None
        self._weight: typing.Optional[ArrayLike] = None
        self.bag_size_flow: typing.Optional[int] = None
        self.bag_size_density: typing.Optional[int] = None

        self.models: models.ModelResults = models.ModelResults()
        self.quantiles: list[float] = []
        pass

    def estimate_model(
        self,
        model_type: str,
        quantiles: typing.Optional[list[float]] = None,
        penalty: typing.Optional[str] = None,
        eta: typing.Optional[float] = None,
        email: str = None,
    ) -> None:
        """ Estimates a convex nonparametric least squares (CNLS) or convex quantile regression (CQR) model for
            bagged data.

        Parameters
        ----------
        model_type : str
            Type of the model to be estimated. Takes value `'quantile'` for the convex quantile regression (wCQR)
            model or `'mean'` for the convex nonparametric least squares (wCNLS) model
        quantiles : list[float]
            Quantiles to be estimated for the `model_type=quantile`
        penalty : str, optional
            Type of penalty to be used in the model. Takes values `penalty=None` (no penalty), `penalty='l1'` (L1 norm),
            `penalty='l2'` (L2 norm), or `penalty='l3'` (Lipschitz norm), by default None
        eta : float, optional
            Value of the tuning parameter if `penalty` is in `['l1', 'l2', 'l3']`, by default None
        email : str, optional
            For external optimization on the NEOS server, by default None

        Raises
        ------
        AssertionError
            Model type must be "quantile" or "mean".
            Quantiles must be a non-empty list.
            Quantiles must be between 0 and 1.
            Data must be loaded first.
            Data contains more than 3000 observations. Apply bagging first.
        """
        # if quantiles is None:
        #    quantiles = [0.5]

        assert model_type in [
            "quantile",
            "mean",
        ], "[LOG] AssertionError: Model type must be 'quantile' or 'mean'."
        assert self.data is not None, "[LOG] AssertionError: No data is available"
        if penalty is not None:
            assert penalty in [
                "l1",
                "l2",
                "l3",
            ], "[LOG] AssertionError:Penalty must be None, 'l1', 'l2' or 'l3'."
            assert (
                eta is not None
            ), "[LOG] AssertionError: eta must be specified if penalty is selected."

        if model_type == "quantile":
            assert (
                len(quantiles) > 0
            ), "[LOG] AssertionError: Quantiles must be a non-empty list."
            assert all(
                [0 < q < 1 for q in quantiles]
            ), "[LOG] AssertionError: Quantiles must be between 0 and 1."

            [self.quantiles.append(x) for x in quantiles if x not in self.quantiles]
            self.quantiles.sort()

            for q in quantiles:
                if penalty is not None:
                    model = models._pwcqr(
                        x=self._density,
                        y=self._flow,
                        weight=self._weight,
                        quantile=q,
                        penalty=penalty,
                        eta=eta,
                        email=email,
                    )
                else:
                    model = models._wcqr(
                        x=self._density, y=self._flow, weight=self._weight, quantile=q, email=email
                    )
                self.models._add_model(model, q, penalty, eta)

        elif model_type == "mean":
            if penalty is not None:
                model = models._pwcnls(
                    x=self._density,
                    y=self._flow,
                    weight=self._weight,
                    penalty=penalty,
                    eta=eta,
                    email=email,
                )
            else:
                model = models._wcnls(x=self._density, y=self._flow, weight=self._weight, email=email)
            self.models._add_model(model, "mean", penalty, eta)
        pass



import typing
import pandas as pd


class RawData:
    """A class to represent raw data of a traffic measurement station."""
    def __init__(self):
        self.data_raw: typing.Optional[pd.DataFrame] = None
        pass



import pandas as pd
import numpy as np
from typing import Sequence
import scipy
import time

from .constants import DEF_AGG_TIME_PER


def aggregate(
    data: pd.DataFrame,
    by_lane: bool = False,
    aggregation_time_period: int = DEF_AGG_TIME_PER,
    aggregation_time_period_unit: str = "min",
) -> pd.DataFrame:
    assert aggregation_time_period_unit in ["min", "sec"], "[LOG]: AssertionError: Wrong aggregation time period unit." # noqa E501
    start_time = time.perf_counter()

    # Copy initial data set to calculate aggregation periods for a selected direction
    df = data.copy(deep=True)
    if aggregation_time_period_unit == "min":
        df["aggregation"] = (df["hour"] * 60 + df["minute"]) / aggregation_time_period
    if aggregation_time_period_unit == "sec":
        df["aggregation"] = (
            df["hour"] * 60 * 60 + df["minute"] * 60 + df["second"]
        ) / aggregation_time_period
    df = df.astype({"aggregation": int})

    # Aggregate flow and speed by time
    if by_lane:
        agg_data = df.groupby(
            ["id", "date", "aggregation", "direction", "lane"], as_index=False
        ).agg(
            period_speed=("speed", scipy.stats.hmean),
            period_flow=("cars", "count"),
            period_cars=("cars", "sum"),
            period_buses=("buses", "sum"),
            period_trucks=("trucks", "sum"),
        )
    else:
        agg_data = df.groupby(
            ["id", "date", "aggregation", "direction"], as_index=False
        ).agg(
            period_speed=("speed", scipy.stats.hmean),
            period_flow=("cars", "count"),
            period_cars=("cars", "sum"),
            period_buses=("buses", "sum"),
            period_trucks=("trucks", "sum"),
        )

    agg_data["flow"] = 60 / aggregation_time_period * agg_data["period_flow"]
    agg_data["cars"] = 60 / aggregation_time_period * agg_data["period_cars"]
    agg_data["buses"] = 60 / aggregation_time_period * agg_data["period_buses"]
    agg_data["trucks"] = 60 / aggregation_time_period * agg_data["period_trucks"]
    agg_data["density"] = agg_data["flow"].div(agg_data["period_speed"].values)
    agg_data["seconds"] = agg_data["aggregation"] * 60 * aggregation_time_period
    agg_data["seconds"] = agg_data["seconds"].astype("float64")
    agg_data["time"] = pd.to_datetime(agg_data["seconds"], unit="s")
    agg_data["time"] = pd.Series([val.strftime("%H:%M") for val in agg_data["time"]])

    end_time = time.perf_counter()

    print(f"Aggregation took {end_time-start_time:.4f} seconds")
    return agg_data


def bagging(
    agg_data: pd.DataFrame, grid_size_x: int = 70, grid_size_y: int = 400
) -> pd.DataFrame:
    """
    This function is used to reduce computational complexity for \
    convex non-parametric modeling. It merges similar data points \
    into centroid ones, which further will be used for estimation.
    Parameters
    ----------
    agg_data : pd.DataFrame
        Aggregated dataframe
    grid_size_x : int, optional
        Number of bags for x-axis, by default 70
    grid_size_y : int, optional
        Number of bags for y-axis, by default 400

    Returns
    -------
    pd.DataFrame
        pd.DataFrame with bagged data
    """
    start_time = time.perf_counter()
    size_before = len(agg_data)

    # Getting the max density and flow values to calculate the size of the bag
    max_density = agg_data["density"].max()
    max_flow = agg_data["flow"].max()

    # Calculating the size of the bag
    grid_density_size = max_density / grid_size_x
    grid_flow_size = max_flow / grid_size_y

    # Assigning the bag number for density and
    agg_data["grid_density"] = agg_data["density"] / grid_density_size
    agg_data["grid_flow"] = agg_data["flow"] / grid_flow_size
    agg_data = agg_data.astype({"grid_density": int, "grid_flow": int})

    # Calculating the centroid and the weight of each bag
    bag_data = agg_data.groupby(
        ["id", "direction", "grid_density", "grid_flow"], as_index=False
    ).agg(
        bag_size=("id", "count"),
        sum_flow=("flow", "sum"),
        sum_density=("density", "sum"),
    )
    bag_data["centroid_flow"] = bag_data["sum_flow"].div(bag_data["bag_size"])
    bag_data["centroid_density"] = bag_data["sum_density"].div(bag_data["bag_size"])
    bag_data["weight"] = bag_data["bag_size"].div(len(agg_data))
    size_after = len(bag_data)

    end_time = time.perf_counter()
    print(
        f"Data bagging took {end_time-start_time:.4f} seconds. Data reduction is {(1 - size_after/size_before)*100.0:.2f}%" # noqa E501
    )

    return bag_data


def representor(
    alpha: Sequence[float], beta: Sequence[float], x: Sequence[float]
) -> Sequence[float]:  # noqa E501
    """
    Parameters
    ----------
    alpha : Sequence[float]
        A numpy array of alpha coefficients
    beta : Sequence[float]
        A numpy array of beta coefficients
    x : Sequence[float]
        A numpy array of input variables
    Calculation of representation function (Kuosmanen, 2008 / Formula 4.1)

    Returns
    -------
    ArrayLike[float]
        The minimum value of g_hat for each x
    """
    alpha_r = alpha.reshape(-1, 1)
    beta_r = beta.reshape(-1, 1)
    x_t = x.reshape(1, -1)
    mult = np.matmul(beta_r, x_t)
    mult_a = alpha_r + mult
    g_hat_min = mult_a.min(axis=0)

    return g_hat_min



# import dependencies
import datetime


def date_to_day(date: datetime.date) -> tuple:
    """
    Changes a date to the day serial number within the year.

    Parameters
    ----------
    date : datetime.time
        A date for which the day serial number is determined.

    Returns
    -------
    int
        A day serial number for the given date
    """
    day = abs(date - datetime.date(date.year, 1, 1)).days + 1
    return int(date.year), day


def day_to_date(year: int, day: int) -> datetime.date:
    """
    Determines the date for the given `day` of the given `year`

    Parameters
    ----------
    year : int
        Year of the day of interest
    day : int
        Day serial number within the `year`

    Returns
    -------
    datetime.date
        A date for the given `day` of the `year`
    """
    date = datetime.date(year, 1, 1) + datetime.timedelta(days=day - 1)
    return date



from setuptools import setup, find_packages

with open("README.md") as readme_file:
    README = readme_file.read()

setup_args = dict(
    name="roadtraffic",
    version="0.0.1",
    description="A python package for the fundamental diagram of \
        traffic flow estimation.",
    python_requires=">=3.8, <4",
    long_description_content_type="text/markdown",
    long_description=README,
    license="GPLv3",
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    author="Iaroslav Kriuchkov",
    author_email="iaroslav.kriuchkov@aalto.fi",
    keywords=["road traffic", "fundamental diagram"],
    url="https://github.com/iarokr/roadtraffic",
    download_url="https://pypi.org/project/roadtraffic/",
    zip_safe=False,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)

install_requires = [
    "pystoned>=0.5.8",
    "pandas>=1.5.2",
    "numpy>=1.24.0",
    "scikit-learn>=1.1.3",
    "scipy>=1.9.3",
    "matplotlib>=3.6.2",
    "requests>=2.28.1",
    "datetime>=4.9",
]

if __name__ == "__main__":
    setup(**setup_args, install_requires=install_requires)



# import dependencies
import typing
import time
from dataclasses import dataclass

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from pystoned import CNLS, pCNLS, wCNLS, pwCNLS, CQER, pCQER, wCQER, pwCQER
from pystoned.constant import RTS_VRS, FUN_PROD, CET_ADDI


@dataclass
class ModelResults:
    """A class to represent estimated for the aggregated data.

    Methods:
    --------
    get_models() -> dict
        Get the dictionary where all the models are stored. The model object is a pystoned model object. Not recommended
        to use this to extract the estimated model parameters. Use the other methods instead.
    get_frontier(quantile: typing.Union[str, float], penalty: typing.Union[None, str] = None, eta: typing.Union[None, float] = None) -> ArrayLike
        Get the array with the estimated values of the piecewise-linear function.
    get_coefficients(quantile: typing.Union[str, float], penalty: typing.Union[None, str] = None, eta: typing.Union[None, float] = None) -> ArrayLike
        Get the array with the estimated coefficients of the piecewise-linear function.
    get_number_of_segments(quantile: typing.Union[str, float], penalty: typing.Union[None, str] = None, eta: typing.Union[None, float] = None) -> int
        Get the number of segments in the estimated piecewise-linear function.
    get_estimate(quantile: typing.Union[str, float], penalty: typing.Union[None, str] = None, eta: typing.Union[None, float] = None) -> pd.DataFrame
        Get all the data for estimated fundamental diagram. Original density, flow and speed values, estimated flow
        and speed values.
    get_problem_status(quantile: typing.Union[str, float], penalty: typing.Union[None, str] = None, eta: typing.Union[None, float] = None) -> int
        Get the status of the optimization problem.
    """  # noqa E501

    def __init__(self):
        self._models = dict()
        pass

    def _add_model(
        self,
        model: typing.Union[
            CNLS.CNLS,
            pCNLS.pCNLS,
            wCNLS.wCNLS,
            pwCNLS.pwCNLS,
            CQER.CQR,
            pCQER.pCQR,
            wCQER.wCQR,
            pwCQER.pwCQR,
        ],
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str],
        eta: typing.Union[None, float],
    ):
        self._models[quantile, penalty, eta] = model
        pass

    def __getitem__(self, item):
        if (item == "mean") or (0 < item < 1):
            try:
                return self._models[item, None, None]
            except KeyError:
                raise KeyError("Model with the specified parameters was not estimated.")
        elif (type(item) is tuple) and (len(item) == 3):
            try:
                return self._models[item]
            except KeyError:
                raise KeyError("Model with the specified parameters was not estimated.")
        else:
            raise KeyError("Model with the specified parameters was not estimated.")
        pass

    def get_models(self):
        """
        Get the dictionary where all the models are stored. The model object is a pystoned model object. Not recommended
        to use this to extract the estimated model parameters. Use the other methods instead.
        """
        return self._models

    def get_frontier(
        self,
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str] = None,
        eta: typing.Union[None, float] = None,
    ) -> ArrayLike:
        """
        Get the array with the estimated values of the piecewise-linear function.
        Parameters
        ----------
        quantile : str or float
            Quantile of the estimated model.
        penalty : str or None, optional
            Penalty type of the estimated model, by default None
        eta : float or None, optional
            Value of the penalty parameter, by default None

        Raises
        ------
        AssertionError
            Model with the specified parameters is not estimated.

        Returns
        -------
        ArrayLike
            Array with the estimated function values.
        """
        assert (
            quantile,
            penalty,
            eta,
        ) in self._models.keys(), (
            "Model with the specified parameters is not estimated."
        )
        return self._models[quantile, penalty, eta].get_frontier()

    def get_coefficients(
        self,
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str] = None,
        eta: typing.Union[None, float] = None,
    ) -> ArrayLike:
        """
        Get the array with the estimated coefficients of the piecewise-linear function.

        Parameters
        ----------
        quantile : str or float
            Quantile of the estimated model.
        penalty : str or None, optional
            Penalty type of the estimated model, by default None
        eta : float or None, optional
            Value of the penalty parameter, by default None

        Raises
        ------
        AssertionError
            Model with the specified parameters is not estimated.

        Returns
        -------
        ArrayLike
            Array with the estimated coefficients, alpha and beta.
        """
        assert (
            quantile,
            penalty,
            eta,
        ) in self._models.keys(), (
            "Model with the specified parameters is not estimated."
        )
        alpha = self._models[quantile, penalty, eta].get_alpha()
        beta = self._models[quantile, penalty, eta].get_beta().flatten()
        # join alpha in beta into one array
        coefficients = (np.stack([alpha, beta], axis=0)).T
        return coefficients

    def get_number_of_segments(
        self,
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str] = None,
        eta: typing.Union[None, float] = None,
    ) -> int:
        """
        Get the number of segments in the estimated piecewise-linear function.
        Parameters
        ----------
        quantile : str or float
            Quantile of the estimated model.
        penalty : str or None, optional
            Penalty type of the estimated model, by default None
        eta : float or None, optional
            Value of the penalty parameter, by default None

        Raises
        ------
        AssertionError
            Model with the specified parameters is not estimated.

        Returns
        -------
        int
            Number of segments in the estimated piecewise-linear function.
        """
        assert (
            quantile,
            penalty,
            eta,
        ) in self._models.keys(), (
            "Model with the specified parameters is not estimated."
        )
        return len(self._models[quantile, penalty, eta].get_alpha())

    def get_estimate(
        self,
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str] = None,
        eta: typing.Union[None, float] = None,
    ) -> pd.DataFrame:
        """Get all the data for estimated fundamental diagram. Original density, flow and speed values, estimated flow
        and speed values.
        Parameters
        ----------
        quantile : str or float
            Quantile of the estimated model.
        penalty : str or None, optional
            Penalty type of the estimated model, by default None
        eta : float or None, optional
            Value of the penalty parameter, by default None

        Raises
        ------
        AssertionError
            Model with the specified parameters is not estimated.

        Returns
        -------
        estimate : pd.DataFrame
            Estimated fundamental diagram values.
        """
        assert (
            quantile,
            penalty,
            eta,
        ) in self._models.keys(), (
            "Model with the specified parameters is not estimated."
        )
        x = np.array(self._models[quantile, penalty, eta].x).flatten().T
        y = np.array(self._models[quantile, penalty, eta].y).T
        y_hat = np.array(self._models[quantile, penalty, eta].get_frontier()).T
        data = (np.stack([x, y, y_hat], axis=0)).T
        data = data[np.argsort(data[:, 0])].T

        x, y, f = data[0], data[1], data[2]
        speed = y / x
        speed_hat = y_hat / x

        estimate = pd.DataFrame(
            {
                "density": x,
                "flow": y,
                "flow_estimate": f,
                "speed": speed,
                "speed_estimate": speed_hat,
            }
        )
        return estimate

    def get_problem_status(
        self,
        quantile: typing.Union[str, float],
        penalty: typing.Union[None, str] = None,
        eta: typing.Union[None, float] = None,
    ) -> int:
        """Get the status of the optimization problem.

        Parameters
        ----------
        quantile : str or float
            Quantile of the estimated model.
        penalty : str or None, optional
            Penalty type of the estimated model, by default None
        eta : float or None, optional
            Value of the penalty parameter, by default None

        Raises
        ------
        AssertionError
            Model with the specified parameters is not estimated.

        Returns
        -------
        int
            Status of the optimization problem.

        """
        assert (
            quantile,
            penalty,
            eta,
        ) in self._models.keys(), (
            "Model with the specified parameters is not estimated."
        )
        return self._models[quantile, penalty, eta].get_status()


# Estimate CNLS model
def _cnls(x: ArrayLike, y: ArrayLike, email: str) -> CNLS.CNLS:
    start_time = time.perf_counter()
    print("[LOG] Estimating the CNLS model...")
    model = CNLS.CNLS(y=y, x=x, fun=FUN_PROD, cet=CET_ADDI, rts=RTS_VRS)
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] CNLS model was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate pCNLS model
def _pcnls(
    x: ArrayLike, y: ArrayLike, penalty: str, eta: float, email: str
) -> pCNLS.pCNLS:
    start_time = time.perf_counter()
    print(f"[LOG] Estimating the CNLS model with {penalty} penalty (eta={eta})...")
    model = pCNLS.pCNLS(
        y=y, x=x, fun=FUN_PROD, cet=CET_ADDI, rts=RTS_VRS, penalty=int(penalty[-1]), eta=eta
    )
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] CNLS model with {penalty} penalty (eta={eta}) was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate wCNLS model
def _wcnls(x: ArrayLike, y: ArrayLike, weight: ArrayLike, email: str) -> wCNLS.wCNLS:
    start_time = time.perf_counter()
    print("[LOG] Estimating the weighted CNLS model...")
    model = wCNLS.wCNLS(y=y, x=x, w=weight, fun=FUN_PROD, cet=CET_ADDI, rts=RTS_VRS)
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] weighted CNLS model was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate pwCNLS model
def _pwcnls(
    x: ArrayLike, y: ArrayLike, weight: ArrayLike, penalty: str, eta: float, email: str
) -> pwCNLS.pwCNLS:
    start_time = time.perf_counter()
    print(
        f"[LOG] Estimating the weighted CNLS model with {penalty} penalty and eta={eta}..."
    )
    model = pwCNLS.pwCNLS(
        y=y,
        x=x,
        w=weight,
        fun=FUN_PROD,
        cet=CET_ADDI,
        rts=RTS_VRS,
        penalty=int(penalty[-1]),
        eta=eta,
    )
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] weighted CNLS model with {penalty} penalty (eta={eta}) was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate CQR model
def _cqr(x: ArrayLike, y: ArrayLike, quantile: float, email: str) -> CQER.CQR:
    start_time = time.perf_counter()
    print(f"[LOG] Estimating the CQR model for quantile={quantile}...")
    model = CQER.CQR(y=y, x=x, tau=quantile, fun=FUN_PROD, cet=CET_ADDI, rts=RTS_VRS)
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] CQR model for quantile={quantile} was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate pCQR model
def _pcqr(
    x: ArrayLike, y: ArrayLike, quantile: float, penalty: str, eta: float, email: str
) -> pCQER.pCQR:
    start_time = time.perf_counter()
    print(
        f"[LOG] Estimating the CQR model for quantile={quantile} with {penalty} penalty (eta={eta})..."
    )
    model = pCQER.pCQR(
        y=y,
        x=x,
        tau=quantile,
        fun=FUN_PROD,
        cet=CET_ADDI,
        rts=RTS_VRS,
        penalty=int(penalty[-1]),
        eta=eta,
    )
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] CQR model for quantile={quantile} with {penalty} penalty (eta={eta}) was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate wCQR model
def _wcqr(
    x: ArrayLike, y: ArrayLike, weight: ArrayLike, quantile: float, email: str
) -> wCQER.wCQR:
    start_time = time.perf_counter()
    print(f"[LOG] Estimating the weighted CQR model for quantile={quantile}...")
    model = wCQER.wCQR(
        y=y, x=x, w=weight, tau=quantile, fun=FUN_PROD, cet=CET_ADDI, rts=RTS_VRS
    )
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] weighted CQR model for quantile={quantile} was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model


# Estimate pwCQR model
def _pwcqr(
    x: ArrayLike,
    y: ArrayLike,
    weight: ArrayLike,
    quantile: float,
    penalty: str,
    eta: float,
    email: str,
) -> pwCQER.pwCQR:
    start_time = time.perf_counter()
    print(
        f"[LOG] Estimating the weighted CQR model for quantile={quantile} with {penalty} penalty (eta={eta})..."
    )
    model = pwCQER.pwCQR(
        y=y,
        x=x,
        w=weight,
        tau=quantile,
        fun=FUN_PROD,
        cet=CET_ADDI,
        rts=RTS_VRS,
        penalty=int(penalty[-1]),
        eta=eta,
    )
    model.__model__.beta.setlb(None)
    if email is not None:
        model.optimize(email)
    else:
        model.optimize()
    end_time = time.perf_counter()
    print(
        f"[LOG] weighted CQR model for quantile={quantile} with {penalty} penalty (eta={eta}) was estimated in {end_time - start_time:.4f} seconds."  # noqa E501
    )
    return model
